@@ -186,6 +186,7 @@ class TestConnectionPool < Minitest::Test
       def mock.disconnect!
         raise "should not disconnect upon explicit return"
       end
+
       mock
     }
 
@@ -246,6 +247,60 @@ class TestConnectionPool < Minitest::Test
     assert_same conn, Thread.new { pool.checkout }.value
   end
 
+  def test_discard
+    pool = ConnectionPool.new(timeout: 0, size: 1) { NetworkConnection.new }
+    pool.checkout
+
+    Thread.new {
+      assert_raises Timeout::Error do
+        pool.checkout
+      end
+    }.join
+
+    pool.discard_current_connection
+    pool.checkin
+
+    assert_equal 1, pool.size
+    assert_equal 0, pool.idle
+    assert_equal 1, pool.available
+  end
+
+  def test_discard_with_argument
+    pool = ConnectionPool.new(timeout: 0, size: 1) { NetworkConnection.new }
+    pool.checkout
+
+    Thread.new {
+      assert_raises Timeout::Error do
+        pool.checkout
+      end
+    }.join
+
+    pool.discard_current_connection { |conn| assert_kind_of NetworkConnection, conn }
+    pool.checkin
+
+    assert_equal 1, pool.size
+    assert_equal 0, pool.idle
+    assert_equal 1, pool.available
+  end
+
+  def test_discard_with_argument_and_error
+    pool = ConnectionPool.new(timeout: 0, size: 1) { NetworkConnection.new }
+    pool.checkout
+
+    Thread.new {
+      assert_raises Timeout::Error do
+        pool.checkout
+      end
+    }.join
+
+    pool.discard_current_connection { |conn| raise "boom" }
+    pool.checkin
+
+    assert_equal 1, pool.size
+    assert_equal 0, pool.idle
+    assert_equal 1, pool.available
+  end
+
   def test_returns_value
     pool = ConnectionPool.new(timeout: 0, size: 1) { Object.new }
     assert_equal 1, pool.with { |o| 1 }
@@ -410,6 +465,33 @@ class TestConnectionPool < Minitest::Test
     assert_equal ["inner", "outer", "other"], recorder.calls
   end
 
+  def test_nested_discard
+    recorder = Recorder.new
+    pool = ConnectionPool.new(size: 1) { {recorder: recorder} }
+    pool.with do |r_outer|
+      @other = Thread.new { |t|
+        pool.with do |r_other|
+          r_other[:recorder].do_work("other")
+        end
+      }
+
+      pool.with do |r_inner|
+        @inner = r_inner
+        r_inner[:recorder].do_work("inner")
+        pool.discard_current_connection
+      end
+
+      Thread.pass
+
+      r_outer[:recorder].do_work("outer")
+    end
+
+    @other.join
+
+    assert_equal ["inner", "outer", "other"], recorder.calls
+    refute_same @inner, pool.checkout
+  end
+
   def test_shutdown_is_executed_for_all_connections
     recorders = []
 
@@ -707,34 +789,25 @@ class TestConnectionPool < Minitest::Test
     assert_predicate(status, :success?)
   end
 
-  def test_ractors
+  def test_ractors_pool_usage
     begin
       Ractor
     rescue NameError
       skip("Ractor not available")
     end
 
-    begin
-      # TODO Ractor prints a bunch of warnings to the console, no idea
-      # how to turn it off
-      r = Ractor.new do
-        ConnectionPool.new(auto_reload_after_fork: true) { Object.new }
-        true
+    silence_warnings do
+      obj = "mike"
+      r = Ractor.new(obj) do |copy|
+        # verify we can create a pool in a Ractor and that we can
+        pool = ConnectionPool.new(auto_reload_after_fork: false) { copy }
+        checkedout = nil
+        pool.with { |y| checkedout = y }
+        checkedout
       end
-      r.take
-      # should not get here
-      refute true
-    rescue Ractor::RemoteError => re
-      # expected
-      assert re.cause
-      assert_equal Ractor::IsolationError, re.cause.class
-      assert_match(/ConnectionPool::INSTANCES/, re.cause.message)
+      result = r.take
+      assert_equal obj, result # same string but different String instance
+      refute_equal obj.object_id, result.object_id # was copied across Ractor boundary
     end
-
-    r = Ractor.new do
-      ConnectionPool.new(auto_reload_after_fork: false) { Object.new }
-      true
-    end
-    assert_equal true, r.take
   end
 end
